//======================================  W-pipe headers
#include "wscan.hh"
#include "wcondition.hh"
#include "weventlist.hh" 
#include "wfigure.hh"
#include "wframecache.hh"
#include "wmeasure.hh"
#include "wreaddata.hh"
#include "wresample.hh"
// #include "wselect.hh"
// #include "wthreshold.hh"
#include "wtile.hh"
#include "wtransform.hh"
#include "matlab_fcs.hh"

//======================================  DMT headers
#include "TSeries.hh"
#include "fSeries/DFT.hh"
#include "Interval.hh"

//======================================  C++ and system headers.
#include <fstream>
#include <sstream>
#include <iostream>
#include <cstdlib>
#include <cmath>
#include <iomanip>
#include <stdexcept>

namespace wpipe {

  inline std::string
  unquote(const std::string& str) {
    if (str.empty()) return str;
    std::string::size_type f = 0;
    std::string::size_type l = str.size();
    if (str[f]   == '\'') ++f;
    if (str[l-1] == '\'') --l;
    return str.substr(f, l-f);
  }

  void wscan(const Time& eventTime, const std::string& configFile, 
	     const std::string& frameCacheFile, const std::string& outDir, 
	     bool generateReport, int debugLevel);
}

using namespace std;
using namespace wpipe;

int 
main(int argc, const char* argv[]) {
  std::string frameCacheFile, outDir;
  bool generateReport = false;
  int debugLevel = 1;
  //for (int i=1; i<argc; i++) {
  //  string argi = argv[i];
  //  if (argi == "-config" ) {
  //    configFile = argv[++i];
  //  }
  //  else if (argi == "-debug-level" ) {
  //    debugLevel = strtol(argv[++i], 0, 0);
  //  }
  //  else if (argi == "-event-time") {
  //    double time = strtod(argv[++i], 0);
  //    long sec = long(time);
  //    eventTime = Time(sec, long((time-sec)*1e9));
  //  }
  //  else if (argi == "-frame-cache" ) {
  //    frameCacheFile = argv[++i];
  //  }
  //  else if (argi == "-generate-report" ) {
  //    generateReport = true;
  //  }
  //  else if (argi == "-out-directory" ) {
  //    outDir = argv[++i];
  //  }
  //  else {
  //    cout << "Unidentified keyword: " << argi << endl;
  //  }
  //}
  if (argc < 3) {
    cerr << "Insufficient arguments" << endl;
  }

  //------------------------------------  Get the start time.
  double time = strtod(argv[1], 0);
  long sec = long(time);
  Time eventTime = Time(sec, long((time-sec)*1e9));

  std::string configFile = argv[2];

  if (argc > 3) frameCacheFile = argv[3];  

  if (argc > 4) outDir = argv[4];

  if (argc > 5) {
    generateReport = strtol(argv[5],0,0) > 0;
  }

  if (argc > 6) {
    debugLevel = strtol(argv[6],0,0);
  }

  try {
    wscan(eventTime, configFile, frameCacheFile, outDir, generateReport, 
	  debugLevel);
  }
  catch (exception& e) {
    cerr << "caught exception: " << e.what() << endl;
  }
}

void
wpipe::wscan(const Time& eventTime, const std::string& configFile, 
      const std::string& frameCacheFile, const std::string& outDir, 
      bool generateReport, int debugLevel) {
  //const double dInf(1.0/0.0);

  wfigure figureHandle;

  /////////////////////////////////////////////////////////////////////////////
  //                                   defaults                              //
  /////////////////////////////////////////////////////////////////////////////

  // default configuration parameters
  double defaultSampleFrequency = 4096;
  double defaultSearchTimeRange = 64;
  double fRange[] = {0, -1};
  dble_vect defaultSearchFrequencyRange(fRange, fRange+2);
  double qRange[] = {4, 64};
  dble_vect defaultSearchQRange(qRange, qRange+2);
  double defaultSearchMaximumEnergyLoss = 0.2;
  double defaultWhiteNoiseFalseRate = 1e-2;
  double defaultSearchWindowDuration = 0.1;
  double plotRange[] = {1, 4, 16};
  dble_vect defaultPlotTimeRanges(plotRange, plotRange+3);
  double plotFRange[]  = {0, -1};
  dble_vect defaultPlotFrequencyRange(plotFRange, plotFRange+2);
  //notused double defaultPlotMaximumEnergyLoss = 0.2;
  double plotNormER[] = {0, 25.5};
  dble_vect defaultPlotNormalizedEnergyRange(plotNormER, plotNormER+2);
  //notused bool defaultAlwaysPlotFlag = 0;

  /////////////////////////////////////////////////////////////////////////////
  //                            hard coded parameters                        //
  /////////////////////////////////////////////////////////////////////////////

  // search parameters
  double transientFactor = 2.0;
  double outlierFactor = 2.0;
  //notused double durationInflation = 1.0;
  //notused double bandwidthInflation = 1.0;

  // display parameters
  int plotHorizontalResolution = 512;
  double plotDurationInflation = 0.5;
  double plotBandwidthInflation = 0.5;

  // limits on number of significant tiles
  int maximumSignificants = 10000;
  int maximumMosaics = 10000;

  /////////////////////////////////////////////////////////////////////////////
  //                        process command line arguments                   //
  /////////////////////////////////////////////////////////////////////////////

  string outputDirectory = outDir;

  // apply default arguments
  string configurationFile = configFile;
  if (configurationFile.empty()) {
    configurationFile = "configuration.txt";
  }

  std::string cacheFile = frameCacheFile;
  if (cacheFile.empty()) {
    cacheFile = "framecache.txt";
  }

  // convert string event time and debug level to numbers
  ostringstream outstr;
  outstr << "0" << fixed << setprecision(9) << eventTime.totalS();
  string eventTimeString = outstr.str();

  // name of html file
  string htmlFile = "index.html";

  // name of text summary file
  string textSummaryFile = "summary.txt";

  // name of xml summary file
  string xmlSummaryFile = "summary.xml";

  // name of context file
  string contextFile = "context.html";

  /////////////////////////////////////////////////////////////////////////////
  //                                 write header                            //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  if (debugLevel >= 1) {
    time_t clock=0;
    cout << "OmegaScan of event at " << eventTimeString << endl;
    cout << "created by " << getenv("USER") << " on " << datestr(clock, 29)
	 <<" at " << datestr(clock, 13) << endl;
  }

  /////////////////////////////////////////////////////////////////////////////
  //                       initialize random number generators               //
  /////////////////////////////////////////////////////////////////////////////

  // set random number generator seeds based on event time
  //rand("state", eventTime);
  //randn("state", eventTime);

  /////////////////////////////////////////////////////////////////////////////
  //                           read configuration file                       //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  if (debugLevel >= 1) {
    cout << "reading configuration file " << configurationFile << endl;
  }
  // enable or disable warnings
  if (debugLevel >= 2) {
    // warning on WSCAN:incompleteConfiguration"
  } else {
    //warning off WSCAN:incompleteConfiguration"
  }

  // initialize configuration structure
  config_vect configuration;
  int channelNumber = 0;
  int sectionNumber = 0;
  section_vect sectList;

  // open configuration file for reading
  ifstream configurationFID(configurationFile.c_str());
  if (!configurationFID.is_open()) {
    throw runtime_error(string("Unable to open configuration file: ")
			+ configurationFile);
  }

  // begin loop over configuration file
  while (!configurationFID.eof()) {

    // read one line from configuration file
    string configurationLine;
    getline(configurationFID, configurationLine);
    
    // remove any comments
    string::size_type commentIndices = configurationLine.find("#");
    if (commentIndices != string::npos) {
      configurationLine.erase(commentIndices);
    }

    // remove leading and trailing blanks
    //configurationLine = fliplr(deblank(fliplr(deblank(configurationLine))));
    configurationLine = deblank(configurationLine);

    // if (empty line, skip to the next line
    if (configurationLine.empty()) {
      continue;
    }

    // check for new section
    if (configurationLine[0] == '[') {

      // locate field separator
      string::size_type commaIndex = configurationLine.find(",");

      // if (field separator not located, report syntax error
      if (commaIndex == string::npos) {
	error(string("syntax error processing configuration file ")
	      + configurationFile + ":\n" + configurationLine);
      }

      // select first field separator
      //commaIndex = commaIndex(1);
      sectList.resize(sectionNumber+1);

      // extract section index
      sectList[sectionNumber].Index=configurationLine.substr(1, commaIndex-1);

      // extract section name
      sectList[sectionNumber].Name=configurationLine.substr(commaIndex+1);

      // remove leading and trailing blanks
      sectList[sectionNumber].Index = deblank(sectList[sectionNumber].Index);
      sectList[sectionNumber].Name = deblank(sectList[sectionNumber].Name);
      sectList[sectionNumber].Name.erase(sectList[sectionNumber].Name.size()-1, 1);

     // standardize section names
      sectList[sectionNumber].Index = strrep(sectList[sectionNumber].Index,
					     ";", ":");
      sectList[sectionNumber].Name = strrep(sectList[sectionNumber].Name, 
					    ";", ":");

      // determine initial visibility
      if (sectList[sectionNumber].Index == "Context") {
	sectList[sectionNumber].Checked = "checked";
	sectList[sectionNumber].Display = "block";
      } else if (sectList[sectionNumber].Index == "Gravitational") {
	sectList[sectionNumber].Checked = "checked";
	sectList[sectionNumber].Display = "block";
      } else {
	sectList[sectionNumber].Checked = "checked";
	sectList[sectionNumber].Display = "block";
	// sectionChecked[sectionNumber] = "";
	// sectionDisplay[sectionNumber] = "none";
      }
      
      // record first channel in section
      sectList[sectionNumber].Start = configuration.size();

      // increment section number
      sectionNumber += 1;

      // continue to next line
      continue;

    }

    // check for beginning of new channel configuration
    if (configurationLine == "{") {

      // increment channel number
      configuration.push_back(search_chan());
      channelNumber = configuration.size() - 1;

      // continue to next line
      continue;

    }

    // check for end of existing channel configuration
    if (configurationLine == "}") {

      // validate channel configuration
      if (configuration[channelNumber].channelName.empty()) {
	cerr << "channel name not specified for channel number " 
	     << channelNumber << endl;
	error("channel name not specified");
      }
      if (configuration[channelNumber].frameType.empty()) {
	cerr << "frame type not specified for channel number " 
	     << channelNumber << endl;
	error("frame type not specified");
      }
      if (configuration[channelNumber].sampleFrequency == 0) {
	cerr << "WSCAN:incompleteConfiguration" 
	     << "sample frequency not specified for channel number"
	     << channelNumber << endl;
	configuration[channelNumber].sampleFrequency = defaultSampleFrequency;
      }
      if (configuration[channelNumber].searchTimeRange == 0) {
	cerr << "WSCAN:incompleteConfiguration"
	     << "search time range not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].searchTimeRange = defaultSearchTimeRange;
      }
      if (fmod(configuration[channelNumber].searchTimeRange, 1) != 0) {
	cerr << "WSCAN:invalidConfiguration: " 
	     << "search time range non integer for channel number "
	     << channelNumber << endl;
	error("WSCAN:invalidConfiguration");
      }
      if (configuration[channelNumber].searchFrequencyRange[1] < 0) {
	cout << "WSCAN:incompleteConfiguration "
	     << "search frequency range not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].searchFrequencyRange[0] = 
	  defaultSearchFrequencyRange[0];
	configuration[channelNumber].searchFrequencyRange[1] = 
	  defaultSearchFrequencyRange[1];
      }
      if (configuration[channelNumber].searchQRange[1] == 0) {
	cout << "WSCAN:incompleteConfiguration "
	     <<	"search Q range not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].searchQRange[0] = defaultSearchQRange[0];
	configuration[channelNumber].searchQRange[1] = defaultSearchQRange[1];
      }
      if (configuration[channelNumber].searchMaximumEnergyLoss == 0) {
	cerr << "WSCAN:incompleteConfiguration"
	     << "search maximum energy loss not specified for channel number"
	     << channelNumber << endl;
	configuration[channelNumber].searchMaximumEnergyLoss = 
	  defaultSearchMaximumEnergyLoss;
      }
      if (configuration[channelNumber].whiteNoiseFalseRate == 0) {
	cerr << "WSCAN:incompleteConfiguration"
	     << "white noise false rate not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].searchMaximumEnergyLoss = 
	  defaultWhiteNoiseFalseRate;
      }
      if (configuration[channelNumber].searchWindowDuration == 0) {
	cerr << "WSCAN:incompleteConfiguration"
	     << "search window duration not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].searchWindowDuration = 
	  defaultSearchWindowDuration;
      }
      if (configuration[channelNumber].plotTimeRanges.empty()) {
	cerr << "WSCAN:incompleteConfiguration" 
	     <<	"plot time ranges not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].plotTimeRanges = defaultPlotTimeRanges;
      }
      if (configuration[channelNumber].plotFrequencyRange.empty()) {
	//cerr << "WSCAN:incompleteConfiguration "
	//     << "plot frequency range not specified for channel number "
	//     << channelNumber << endl;
	configuration[channelNumber].plotFrequencyRange = 
	  defaultPlotFrequencyRange;
      }
      if (configuration[channelNumber].plotNormalizedEnergyRange.empty()) {
	cerr << "WSCAN:incompleteConfiguration "
	     <<"plot normalized energy range not specified for channel number "
	     << channelNumber << endl;
	configuration[channelNumber].plotNormalizedEnergyRange = 
	  defaultPlotNormalizedEnergyRange;
      }
      //if (configuration[channelNumber].alwaysPlotFlag == false) {
      //cerr << "WSCAN:incompleteConfiguration"
      //     << "always plot flag not specified for channel number "
      //     << channelNumber << endl;
      //configuration[channelNumber].alwaysPlotFlag = defaultAlwaysPlotFlag;
      //}

      // continue to next line
      continue;
    }

    // locate field separator
    string::size_type colonIndex = configurationLine.find(":");

    // if (field separator not located, report syntax error
    if (colonIndex == string::npos) {
      cerr << "syntax error processing configuration file "
	   << configurationFile << ":" << endl
	   << configurationLine << endl;
      error("syntax error processing configuration file");
    }
    
    // parse configuration line
    string parameterName = configurationLine.substr(0, colonIndex);
    string parameterValue = configurationLine.substr(colonIndex + 1);
    parameterName = deblank(parameterName);
    parameterValue = deblank(parameterValue);
    const char* parValue = parameterValue.c_str();

    // assign parameters based on name
    if (parameterName == "channelName") {
      configuration[channelNumber].channelName = unquote(parameterValue);
    } else if (parameterName == "frameType") {
      configuration[channelNumber].frameType = unquote(parameterValue);
    } else if (parameterName ==  "sampleFrequency") {
      configuration[channelNumber].sampleFrequency = strtod(parValue, 0);
    } else if (parameterName ==  "searchTimeRange") {
      configuration[channelNumber].searchTimeRange = strtod(parValue, 0);
    } else if (parameterName ==  "searchFrequencyRange") {
      configuration[channelNumber].searchFrequencyRange = eval(parameterValue);
    } else if (parameterName ==  "searchQRange") {
      configuration[channelNumber].searchQRange = eval(parameterValue);
    } else if (parameterName ==  "searchMaximumEnergyLoss") {
      configuration[channelNumber].searchMaximumEnergyLoss =strtod(parValue, 0);
    } else if (parameterName ==  "whiteNoiseFalseRate") {
      configuration[channelNumber].whiteNoiseFalseRate = strtod(parValue, 0);
    } else if (parameterName ==  "searchWindowDuration") {
      configuration[channelNumber].searchWindowDuration = strtod(parValue, 0);
    } else if (parameterName ==  "plotTimeRanges") {
      configuration[channelNumber].plotTimeRanges = eval(parameterValue);
    } else if (parameterName ==  "plotFrequencyRange") {
      configuration[channelNumber].plotFrequencyRange = eval(parameterValue);
    } else if (parameterName ==  "plotNormalizedEnergyRange") {
      configuration[channelNumber].plotNormalizedEnergyRange = 
	eval(parameterValue);
    } else if (parameterName ==  "alwaysPlotFlag") {
      if (parameterValue == "true") {
	configuration[channelNumber].alwaysPlotFlag = true;
      } else if (parameterValue == "false") {
	configuration[channelNumber].alwaysPlotFlag = false;
      } else {
	configuration[channelNumber].alwaysPlotFlag = strtol(parValue,0,0)!=0;
      }
    } else {
      cerr << "unknown configuration parameter " << parameterName << endl;
      error("unknown configuration parameter");
    }

    // end loop over channel configuration file
  }

  // close configuration file
  configurationFID.close();

  // number of configured channels
  int numberOfChannels = configuration.size();

  // number of sections
  int numberOfSections = sectList.size();

  /////////////////////////////////////////////////////////////////////////////
  //                           load frame cache file                         //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  if (debugLevel >= 1) {
    cout << "reading frame cache file " << cacheFile << endl;
  }

  // load frame file cache
  wframecache frameCache(cacheFile);

  /////////////////////////////////////////////////////////////////////////////
  //                         create output directory                         //
  /////////////////////////////////////////////////////////////////////////////

  // if (outputDirectory not specified, make one based on center time
  if (outputDirectory.empty()) {
    outputDirectory = string("scans/") + eventTimeString;
  }

  // report status
  if (debugLevel >= 1) {
    cout << "creating event directory" << endl;
    cout << "  outputDirectory:         " << outputDirectory << endl;
  }

  // create spectrogram directory
  string shell_cmd;
  shell_cmd  = string("mkdir -p ") + outputDirectory + ";";
  shell_cmd += string("cp ") + configurationFile + " " + outputDirectory 
    + "/configuration.txt";

  // copy configuration file
  if (system(shell_cmd.c_str())) {
    error(string("shell command: \"") + shell_cmd + " \"failed");
  }

  /////////////////////////////////////////////////////////////////////////////
  //                      initialize text summary report                     //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  wlog(debugLevel, 1, "opening text summary");

  // open text summary file
  string fullTextSummaryFile = outputDirectory + "/" + textSummaryFile;
  FILE* textSummaryFID = fopen(fullTextSummaryFile.c_str(), "w");

  // write column definitions
  fprintf(textSummaryFID, "# channelName\n");
  fprintf(textSummaryFID, "# peakTime\n");
  fprintf(textSummaryFID, "# peakFrequency\n");
  fprintf(textSummaryFID, "# peakQ\n");
  fprintf(textSummaryFID, "# peakSignificance\n");
  fprintf(textSummaryFID, "# peakAmplitude\n");

  /////////////////////////////////////////////////////////////////////////////
  //                      initialize xml summary report                      //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  wlog(debugLevel, 1, "opening xml summary");

  // open xml summary file
  string fullXmlSummary = outputDirectory +  "/" + xmlSummaryFile;
  FILE* xmlSummaryFID = fopen(fullXmlSummary.c_str(), "w");

  // write channel definitions
  fprintf(xmlSummaryFID, "<?xml version='1.0' encoding='utf-8' ?>\n");
  fprintf(xmlSummaryFID, "<LIGO_LW>\n");
  fprintf(xmlSummaryFID, "<Table Name=\"wscan:summary:table\">\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:channelName\" Type=\"lstring\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:peakTime\" Type=\"real_8\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:peakFrequency\" Type=\"real_8\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:peakQ\" Type=\"real_8\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:peakSignificance\" Type=\"real_8\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Column Name=\"wscan:summary:peakAmplitude\" Type=\"real_8\"/>\n");
  fprintf(xmlSummaryFID, 
	  "<Stream Name=\"wscan:summary:table\" Type=\"Local\" Delimiter=\",\">\n");

  //////////////////////////////////////////////////////////////////////////////
  //                           initialize html report                         //
  //////////////////////////////////////////////////////////////////////////////

  FILE* htmlFID = NULL;
  if (generateReport) {

    // report status
    wlog(debugLevel, 1, "opening html report");

    // open web page
    string fullHtmlFile = outputDirectory + "/" + htmlFile;
    htmlFID = fopen(fullHtmlFile.c_str(), "w");

    // begin html
    fprintf(htmlFID, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
    fprintf(htmlFID, "<html>\n");
    fprintf(htmlFID, "<head>\n");
    fprintf(htmlFID, "<title>Scan %s</title>\n", eventTimeString.c_str());
    fprintf(htmlFID, "<link rel=\"icon\" type=\"image/x-icon\" href=\"omega.logo.icon.png\" />\n");
    fprintf(htmlFID, "<link rel=\"stylesheet\" type=\"text/css\" href=\"wstyle.css\" />\n");
    fprintf(htmlFID, "<script type=\"text/javascript\">\n");
    fprintf(htmlFID, "function showImage(gpsTime, channelName, timeRanges, imageType) {\n");
    fprintf(htmlFID, "  for (var timeRangeIndex in timeRanges) {\n");
    fprintf(htmlFID, "    var imageBaseName =\n");
    fprintf(htmlFID, "      gpsTime + \"_\" + channelName + \"_\" + timeRanges[timeRangeIndex];\n");
    fprintf(htmlFID, "    document.getElementById(\"a_\" + imageBaseName).href =\n");
    fprintf(htmlFID, "      imageBaseName + \"_\" + imageType + \".png\";\n");
    fprintf(htmlFID, "    document.getElementById(\"img_\" + imageBaseName).src =\n");
    fprintf(htmlFID, "      imageBaseName + \"_\" + imageType + \".thumb.png\";\n");
    fprintf(htmlFID, "  }\n");
    fprintf(htmlFID, "}\n");
    fprintf(htmlFID, "function toggleVisible(division) {\n");
    fprintf(htmlFID, "  if (document.getElementById(\"div_\" + division).style.display == \"none\") {\n");
    fprintf(htmlFID, "    document.getElementById(\"div_\" + division).style.display = \"block\";\n");
    fprintf(htmlFID, "    document.getElementById(\"input_\" + division).checked = true;\n");
    fprintf(htmlFID, "  } else {\n");
    fprintf(htmlFID, "    document.getElementById(\"div_\" + division).style.display = \"none\";\n");
    fprintf(htmlFID, "    document.getElementById(\"input_\" + division).checked = false;\n");
    fprintf(htmlFID, "  } \n");
    fprintf(htmlFID, "}\n");
    fprintf(htmlFID, "function gotoSection(section) {\n");
    fprintf(htmlFID, "  document.getElementById(\"div_\" + section).style.display = \"block\";\n");
    fprintf(htmlFID, "  document.getElementById(\"input_\" + section).checked = true;\n");
    fprintf(htmlFID, "  window.location.hash = section;\n");
    fprintf(htmlFID, "}\n");
    fprintf(htmlFID, "</script>\n");
    fprintf(htmlFID, "</head>\n");
    fprintf(htmlFID, "<body>\n");
    fprintf(htmlFID, "\n");
  
    fprintf(htmlFID, "<h1 class=\"title\"><a href=\"https://geco.phys.columbia.edu/omega\" class=\"title\">Omega Pipeline</a> Scan</h1>\n");
    fprintf(htmlFID, "<h1>Scan %s</h1>\n", eventTimeString.c_str());
    fprintf(htmlFID, "\n");
    fprintf(htmlFID, "<div class=\"main\">\n");
    fprintf(htmlFID, "\n");

    /////////////////////////////////////////////////////////////////////
    //                           add index to html report              //
    /////////////////////////////////////////////////////////////////////

    fprintf(htmlFID, "<h2>Index</h2>\n");
    fprintf(htmlFID, "<div class=\"section\">\n");
    fprintf(htmlFID, "<ul>\n");
    for (sectionNumber=0; 
	 sectionNumber<numberOfSections; 
	 sectionNumber++) {
      fprintf(htmlFID, 
	      "<li><a href=\"javascript:gotoSection('%s')\">%s</a></li>\n", 
	      sectList[sectionNumber].Index.c_str(), 
	      sectList[sectionNumber].Index.c_str());
    }
    fprintf(htmlFID, "<li><a href=\"log.txt\">Scan Log</a></li>\n");
    fprintf(htmlFID, "<li><a href=\"%s\">Text Summary</a></li>\n", 
	    textSummaryFile.c_str());
    fprintf(htmlFID, "<li><a href=\"%s\">XML Summary</a></li>\n", 
	    xmlSummaryFile.c_str());
    fprintf(htmlFID, "<li><a href=\"configuration.txt\">Configuration</a></li>\n");
    fprintf(htmlFID, "</ul>\n");
    fprintf(htmlFID, "</div>\n");
    fprintf(htmlFID, "\n");
	  
    ///////////////////////////////////////////////////////////////////////////
    //                            initialize figure handle                   //
    ///////////////////////////////////////////////////////////////////////////

    figureHandle.open();
	  
  } // generateReport

    ///////////////////////////////////////////////////////////////////////////
    //                       loop over configuration channels                //
    ///////////////////////////////////////////////////////////////////////////
  
  resampler resample;

    // initialize section status
  int inSectionFlag = 0;
  
  // begin loop over channel configurations
  for (int channelNumber=0; channelNumber<numberOfChannels; channelNumber++) {
    dble_vect emptyRange;

    ///////////////////////////////////////////////////////////////////////////
    //                        add section to html report                     //
    ///////////////////////////////////////////////////////////////////////////

    if (generateReport) {

      // begin loop over new sections
      for (int sectionNumber=0; 
	   sectionNumber < numberOfSections; 
	   sectionNumber++) {

	if (sectList[sectionNumber].Start > channelNumber) break;
	if (sectList[sectionNumber].Start != channelNumber) continue;
	  
	// end any previously existing section
	if (inSectionFlag == 1) {
	  fprintf(htmlFID, "</div>\n");
	  fprintf(htmlFID, "\n");
	}

	// label section
	fprintf(htmlFID, "<a name=\"%s\"></a>\n", 
		sectList[sectionNumber].Index.c_str());
	fprintf(htmlFID, "<h2 class=\"%s\">\n", 
		sectList[sectionNumber].Index.c_str());

	fprintf(htmlFID, "<input id=\"input_%s\" type=\"checkbox\" %s\n", 
		sectList[sectionNumber].Index.c_str(), 
		sectList[sectionNumber].Checked.c_str());
	fprintf(htmlFID, "       onclick=\"toggleVisible('%s');\" />\n", 
		sectList[sectionNumber].Index.c_str());
	fprintf(htmlFID, "%s\n", sectList[sectionNumber].Name.c_str());
	fprintf(htmlFID, "</h2>\n");
	fprintf(htmlFID, "\n");
	fprintf(htmlFID, 
		"<div class=\"section\" id=\"div_%s\" style=\"display: %s;\">\n", 
		sectList[sectionNumber].Index.c_str(), 
		sectList[sectionNumber].Display.c_str());
	fprintf(htmlFID, "\n");

	// record start of section
	inSectionFlag = 1;

	// transcribe context information
	if (sectList[sectionNumber].Index == "Context" ||
	    sectList[sectionNumber].Index == "Timing") {
	  contextFile = outputDirectory + "/" + contextFile;
	  if (exist(contextFile,"file") ) {
	    wlog(debugLevel, 1, "  adding context file");
	    
	    fprintf(htmlFID, "<!-- BEGIN CONTEXT -->\n");
	    fprintf(htmlFID, "\n");
	    ifstream ctxstream(contextFile.c_str());
	    string ctxstring;
	    while (!ctxstream.eof()) {
	      getline(ctxstream, ctxstring);
	      fprintf(htmlFID, "%s\n", ctxstring.c_str());
	    }
	    ctxstream.close();
	    fprintf(htmlFID, "\n");
	    fprintf(htmlFID, "<!-- END CONTEXT -->\n");
	    fprintf(htmlFID, "\n");
	  } else {
	    wlog(debugLevel, 1, "  context file not found. skipping");
	  }
	}  else if (sectList[sectionNumber].Index == "Parameters") {
	  fprintf(htmlFID, "<!-- BEGIN PARAMETERS -->\n");
	  fprintf(htmlFID, "\n");
	  fprintf(htmlFID, "<!-- END PARAMETERS -->\n");
	  fprintf(htmlFID, "\n");
	} else if (sectList[sectionNumber].Index == "Notes") {
	  fprintf(htmlFID, "<!-- BEGIN NOTES -->\n");
	  fprintf(htmlFID, "\n");
	  fprintf(htmlFID, "<!-- END NOTES -->\n");
	  fprintf(htmlFID, "\n");
	}

      }	// end loop over new sections

    } // generateReport
  
    /////////////////////////////////////////////////////////////////////////
    //               identify statistically significant channels           //
    /////////////////////////////////////////////////////////////////////////

    // extract search configuration for channel
    string channelNameString = configuration[channelNumber].channelName;
    string frameType = configuration[channelNumber].frameType;
    double sampleFrequency = configuration[channelNumber].sampleFrequency;
    double timeRange = configuration[channelNumber].searchTimeRange;
    dble_vect frequencyRange=configuration[channelNumber].searchFrequencyRange;
    dble_vect qRange = configuration[channelNumber].searchQRange;
    double maximumEnergyLoss = configuration[channelNumber].searchMaximumEnergyLoss;
    double searchWindowDuration = configuration[channelNumber].searchWindowDuration;
    double whiteNoiseFalseRate = configuration[channelNumber].whiteNoiseFalseRate;
    bool alwaysPlotFlag = configuration[channelNumber].alwaysPlotFlag;

    // standardize channel names
    channelNameString = strrep(channelNameString, ";", ":");

    // Plot parameters
    dble_vect plotTimeRanges;
    dble_vect plotFrequencyRange;
    dble_vect plotNormalizedEnergyRange;
    if (generateReport) {
      plotTimeRanges = configuration[channelNumber].plotTimeRanges;
      plotFrequencyRange = configuration[channelNumber].plotFrequencyRange;
      plotNormalizedEnergyRange = 
	configuration[channelNumber].plotNormalizedEnergyRange;
    }

    // display status
    if (debugLevel >= 1) {
      cout << "processing channel " << channelNameString << endl;
    }

    // censor channels used for "blind" injections
    if (channelNameString == "H1:LSC-ETMY_EXC_DAQ" ||
	channelNameString == "H2:LSC-ETMY_EXC_DAQ" || 
	channelNameString == "L1:LSC-ETMY_EXC_DAQ") {
      if (debugLevel >= 1) {
	cout << "  WARNING: " << channelNameString 
	     << " censored for 'blind' injections." << endl;
      }
      continue;
    }

    // override censor for channels used in "blind" injections
    if (channelNameString == "H1:LSC-ETMY_EXC_DAQ!" || 
	channelNameString == "H2:LSC-ETMY_EXC_DAQ!" ||
	channelNameString == "L1:LSC-ETMY_EXC_DAQ!") {
      channelNameString = strrep(channelNameString, "!", "");
      if (debugLevel >= 1) {
	cout << "  WARNING: overriding censor for " << channelNameString 
	     << "." << endl;
      }
    }
  
    // find closest sample time to event time
    Time centerTime = Time(eventTime.getS());
    long sampleOff = long(double(eventTime - centerTime)*sampleFrequency + 0.5);
    centerTime += double(sampleOff) / sampleFrequency;

    // determine segment start and stop times
    Time startTime = centerTime - Interval(timeRange / 2.);
    if (startTime-Time(startTime.getS()) < Interval(0.5)) {
      startTime = Time(startTime.getS());
    } else {
      startTime = Time(startTime.getS()+1);
    }
    Time stopTime = startTime + timeRange;

    // generate search tiling
    wlog(debugLevel, 2, "  tiling for search\n");
    double highPassCutoff = -1;
    double lowPassCutoff = -1;
    double whiteningDuration = -1;

    wtile tiling(timeRange, qRange, frequencyRange, sampleFrequency, 
		 maximumEnergyLoss, highPassCutoff, lowPassCutoff, 
		 whiteningDuration, transientFactor);


    //=========================================================================
    //  This has to be fixed for reasonable data I?O
    // read data from frame file
    wlog(debugLevel, 2, "  reading data\n");
    double timeShifts = 0;
    tser_vect rawData;
    wreaddata(frameCache, channelNameString, frameType, 
	      startTime, stopTime, timeShifts, debugLevel,
	      rawData);

    // check for read error
    if (rawData.empty()) {
      wlog(debugLevel, 1, "  ERROR: cannot load frame data");
      continue;
    }

    // test multiple channels for consistency
    if (rawData.size() > 1) {
      wlog(debugLevel, 2, "  subtracting data\n");
      rawData[0] -= rawData[1];
      rawData.erase(rawData.begin()+1, rawData.end());
    }

    // check for all zeros
    if (rawData[0].getMinimum() == rawData[0].getMaximum()) {
      wlog(debugLevel, 1, "  WARNING: channel contains no information");
      continue;
    }

    //====================================================================

    // resample data
    wlog(debugLevel, 2, "  resampling data\n");
    tser_vect rsamp;
    rsamp = resample.wresample(rawData, sampleFrequency);

    // high pass filter and whiten data
    wlog(debugLevel, 2, "  conditioning data\n");

    wcondition conditioner(rsamp, tiling);
    containers::DFT whitenedData = conditioner.whitenedDFT();

    // q transform whitened data
    wlog(debugLevel, 2, "  transforming whitened data\n");
    wtransform whitenedTransform(whitenedData, tiling, outlierFactor, 
				 channelNameString);
    //whitenedTransform.dump(cout);

    // identify most significant whitened transform tile
    wlog(debugLevel, 2, "  measuring peak significance\n");
    Time thresholdReferenceTime = centerTime;
    dble_vect thresholdTimeRange(2);
    thresholdTimeRange[0] = -0.5*searchWindowDuration;
    thresholdTimeRange[1] =  0.5*searchWindowDuration;
    dble_vect thresholdQRange;
    dble_vect thresholdFrequencyRange;
    wmeasure whitenedProperties(whitenedTransform, tiling, startTime, 
				thresholdReferenceTime, thresholdTimeRange, 
				thresholdFrequencyRange, thresholdQRange, 
				debugLevel);
    //whitenedProperties.dump(cout);
    
    // determine normalized energy threshold
    double normalizedEnergyThreshold = 
      -log(whiteNoiseFalseRate * tiling.duration() / 
	   (1.5 * tiling.numberOfIndependents()));

    // if (channel not significant
    if (whitenedProperties[0].peak.NormalizedEnergy < 
	normalizedEnergyThreshold) {

      // if (always plot flag is not set
      if (alwaysPlotFlag == 0) {

	// report status and skip to the next channel
	if (debugLevel >= 1) {
	  cout << "  channel not significant at white noise false rate " 
	       << whiteNoiseFalseRate << endl;
	}
	fprintf(textSummaryFID, 
		"%-30s %#020.9f %#09.3e %#09.3e %#09.3e %#09.3e\n", 
		channelNameString.c_str(), 0.0, 0.0, 0.0, 0.0, 0.0);
	fprintf(xmlSummaryFID, 
		"\"%s\",%#020.9f,%#09.3e,%#09.3f,%#09.3e,%#09.3e,\n", 
		channelNameString.c_str(), 0.0, 0.0, 0.0, 0.0, 0.0);
	continue;

	// end test for always plot flag
      }

      // end test of channel significance
    }

    ///////////////////////////////////////////////////////////////////////////
    //                analyze statistically significant channels             //
    ///////////////////////////////////////////////////////////////////////////
  
    // q transform raw data
    wlog(debugLevel, 2, "  transforming raw data\n");
    wtransform rawTransform(conditioner.rawDFT(), tiling, outlierFactor, 
			    channelNameString);
    //rawTransform.dump(cout);

    weventstack rawEvents, whitenedEvents;
    if (generateReport) {

      // identify significant whitened q transform tiles
      wlog(debugLevel, 2, "  thresholding whitened transform\n");
      double maxPlotTRange = plotTimeRanges[0];
      for (size_t i=1; i<plotTimeRanges.size(); i++) {
	if (maxPlotTRange < plotTimeRanges[i]) 
	  maxPlotTRange = plotTimeRanges[i];
      }
      thresholdTimeRange[0] = -0.5 * (maxPlotTRange 
				      + tiling.end_plane().row(0).duration 
				      * plotDurationInflation);
      thresholdTimeRange[1] = -thresholdTimeRange[0];
      thresholdFrequencyRange = plotFrequencyRange;
      thresholdQRange = emptyRange;
      weventstack whitenedSignificants;
      whitenedSignificants.wthreshold(
		 whitenedTransform, tiling, startTime, whiteNoiseFalseRate, 
		 thresholdReferenceTime, thresholdTimeRange, 
		 thresholdFrequencyRange, thresholdQRange, 
		 maximumSignificants, "", 0, 0, 0, 
		 debugLevel);
      //whitenedSignificants.dump(cout);

      // identify significant raw q transform tiles
      wlog(debugLevel, 2, "  thresholding raw transform\n");
      weventstack rawSignificants;
      rawSignificants.wthreshold(rawTransform, tiling, startTime, 
				 whiteNoiseFalseRate, thresholdReferenceTime, 
				 thresholdTimeRange, thresholdFrequencyRange, 
				 thresholdQRange, maximumSignificants, "", 0, 
				 0, 0, debugLevel);
      //rawSignificants.dump(cout);

      // select non-overlapping whitened significant tiles
      wlog(debugLevel, 2, "  selecting whitened events\n");
      whitenedEvents.wselect(whitenedSignificants, plotDurationInflation, 
			     plotBandwidthInflation,  maximumMosaics, 
			     debugLevel);

      // select non-overlapping raw significant tiles
      wlog(debugLevel, 2, "  selecting raw events\n");
      rawEvents.wselect(rawSignificants, plotDurationInflation, 
			plotBandwidthInflation, maximumMosaics, debugLevel);

      // recover time series data
      //rawData = wifft(rawData);
      //highPassedData = wifft(highPassedData);
      //whitenedData = wifft(whitenedData);

    } // generateReport

    ///////////////////////////////////////////////////////////////////////////
    //                      add channel to summary report                    //
    ///////////////////////////////////////////////////////////////////////////

    // identify most significant raw transform tile
    wlog(debugLevel, 2, "  measuring peak properties\n");
    thresholdReferenceTime = whitenedProperties[0].peak.centerTime();
    thresholdTimeRange[0] = -whitenedProperties[0].peak.Duration;
    thresholdTimeRange[1] = -thresholdTimeRange[0];
    thresholdFrequencyRange.resize(2);
    thresholdFrequencyRange[0] = whitenedProperties[0].peak.Frequency 
                               - whitenedProperties[0].peak.Bandwidth;
    thresholdFrequencyRange[1] = whitenedProperties[0].peak.Frequency 
                               + whitenedProperties[0].peak.Bandwidth;
    thresholdQRange = emptyRange;
    wmeasure rawProperties(rawTransform, tiling, startTime, 
			   thresholdReferenceTime, thresholdTimeRange, 
			   thresholdFrequencyRange, thresholdQRange, 
			   debugLevel);
    //rawProperties.dump(cout);
 
    // most significant tile properties
    Time mostSignificantTime =  whitenedProperties[0].peak.centerTime();
    double mostSignificantFrequency = whitenedProperties[0].peak.Frequency;
    double mostSignificantQ =  whitenedProperties[0].peak.Q;
    double mostSignificantNormalizedEnergy =
      whitenedProperties[0].peak.NormalizedEnergy;

    //----------------------------------------------------------------------
    //   This was changed to use signalAmplitude[0]
    // double mostSignificantAmplitude = rawProperties{1}.peakAmplitude;
    // double mostSignificantAmplitude = rawProperties[0].signalAmplitude;
    double mostSignificantAmplitude = rawProperties[0].signal.Amplitude;

    // write most significant tile properties to text summary file
    wlog(debugLevel, 2, "  writing text summary\n");
    fprintf(textSummaryFID, 
	    "%-30s %#020.9f %#09.3e %#09.3e %#09.3e %#09.3e\n", 
	    channelNameString.c_str(), mostSignificantTime.totalS(), 
	    mostSignificantFrequency, mostSignificantQ, 
	    mostSignificantNormalizedEnergy, mostSignificantAmplitude);
    fflush(textSummaryFID);

    // write most significant tile properties to xml summary file
    wlog(debugLevel, 2, "  writing xml summary\n");
    fprintf(xmlSummaryFID, 
	    "\"%s\",%#020.9f,%#09.3e,%#09.3e,%#09.3e,%#09.3e,\n", 
	    channelNameString.c_str(), mostSignificantTime.totalS(), 
	    mostSignificantFrequency, mostSignificantQ, 
	    mostSignificantNormalizedEnergy, mostSignificantAmplitude);

    ///////////////////////////////////////////////////////////////////////////
    //                        add channel to html report                     //
    ///////////////////////////////////////////////////////////////////////////

    if (generateReport) {
      ostringstream oss;

      // html for most significant tile properties
      oss << fixed << setprecision(3) << mostSignificantTime.totalS();
      string mostSignificantTimeHtml = oss.str();
      oss.str("");

      oss << fixed << setprecision(1) << mostSignificantFrequency;
      string mostSignificantFrequencyHtml =  oss.str();
      oss.str("");

      oss << fixed << setprecision(1) << mostSignificantQ;
      string  mostSignificantQHtml =   oss.str();
      oss.str("");

      double dexp = log10(mostSignificantNormalizedEnergy);
      int expo = int(dexp);
      if (dexp < 0) expo--;
      oss << fixed << setprecision(1) 
	  << mostSignificantNormalizedEnergy/pow(10.0, expo)
	  << "&times;10<sup>" << expo << "</sup>";
      string mostSignificantNormalizedEnergyHtml =  oss.str();
      oss.str("");

      dexp = log10(mostSignificantAmplitude);
      expo = int(dexp);
      if (dexp < 0) expo--;
      oss << fixed << setprecision(1) 
	  << mostSignificantAmplitude / pow(10.0, expo)
	  << "&times;10<sup>" << expo << "</sup>";
      string mostSignificantAmplitudeHtml =  oss.str();
      oss.str("");

      oss << fixed << setprecision(1) 
	  << sqrt(2 * mostSignificantNormalizedEnergy);
      string mostSignificantSNRHtml =  oss.str();
      oss.str("");

      // begin html channel entry
      wlog(debugLevel, 2, "  writing html report\n");
      fprintf(htmlFID, "<a name=\"%s\"></a>\n", channelNameString.c_str());
      fprintf(htmlFID, "<h3>\n");
      fprintf(htmlFID, "<input id=\"input_%s\" type=\"checkbox\" checked\n", 
	      channelNameString.c_str());
      fprintf(htmlFID, "       onclick=\"toggleVisible('%s');\" />\n", 
	      channelNameString.c_str());
      fprintf(htmlFID, "<a href=\"http://ldas-jobs.ligo.caltech.edu/cgi-bin/chanwiki?%s\">\n", 
	      channelNameString.c_str());
      fprintf(htmlFID, "%s\n", channelNameString.c_str());
      fprintf(htmlFID, "</a>\n");
      fprintf(htmlFID, "</h3>\n");

      fprintf(htmlFID, "<div id=\"div_%s\" style=\"display: block;\">\n", 
	      channelNameString.c_str());

      fprintf(htmlFID, "most significant tile:\n" 
	      " t = %s s,\n" 
	      " f = %s Hz,\n" 
	      " Q = %s,\n" 
	      " Z = %s,\n" 
	      " X = %s Hz<sup>-&frac12;</sup>,\n" 
	      " SNR = %s\n"
	      "<br />\n", 
	      mostSignificantTimeHtml.c_str(), 
	      mostSignificantFrequencyHtml.c_str(), 
	      mostSignificantQHtml.c_str(), 
	      mostSignificantNormalizedEnergyHtml.c_str(), 
	      mostSignificantAmplitudeHtml.c_str(), 
	      mostSignificantSNRHtml.c_str());

      oss << "'" << fixed << setprecision(2) << plotTimeRanges[0] << "'";
      for (size_t i=1; i < plotTimeRanges.size(); i++) {
	oss << ", '" << fixed << setprecision(2) << plotTimeRanges[i] << "'";
      }
      string timeRangeString = oss.str();

      fprintf(htmlFID, "  time series:\n");
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'timeseries_raw');\">raw</a>,\n", 
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'timeseries_highpassed');\">high passed</a>,\n",
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'timeseries_whitened');\">whitened</a>\n",
	      timeRangeString.c_str());
      fprintf(htmlFID, "| spectrogram:\n");
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'spectrogram_raw');\">raw</a>,\n", 
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'spectrogram_whitened');\">whitened</a>,\n", 
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'spectrogram_autoscaled');\">autoscaled</a>\n",
	      timeRangeString.c_str());
      fprintf(htmlFID, "| eventgram:\n");
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'eventgram_raw');\">raw</a>,\n", 
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'eventgram_whitened');\">whitened</a>,\n",
	      timeRangeString.c_str());
      fprintf(htmlFID, "<a href=\"javascript:showImage('%s', '%s',\n",
	      eventTimeString.c_str(), channelNameString.c_str());
      fprintf(htmlFID, "[%s], 'eventgram_autoscaled');\">autoscaled</a>\n",
	      timeRangeString.c_str());
      fprintf(htmlFID, "<br />\n");

      /////////////////////////////////////////////////////////////////////////
      //                       begin loop over time ranges                   //
      /////////////////////////////////////////////////////////////////////////

      // begin loop over time ranges
      for (size_t iPlTiR=0; iPlTiR < plotTimeRanges.size(); iPlTiR++) {
	double plotTimeRange =  plotTimeRanges[iPlTiR];
	dble_vect autoNormalizedEnergyRange;

	// report status
	if (debugLevel >= 2) {
	  cout << "  processing" << plotTimeRange << " second time range" 
	       << endl;
	}

	// determine start and stop times of figures
	//notused double plotStartTime = timeRange / 2 - plotTimeRange / 2;
	//notused double plotStopTime = timeRange / 2 + plotTimeRange / 2;

	///////////////////////////////////////////////////////////////////////
	//                         plot raw time series                      //
	///////////////////////////////////////////////////////////////////////

	// plot raw time series
	wlog(debugLevel, 2, "    plotting raw time series\n");
	figureHandle.clear();
	figureHandle.wtimeseries(rawData[0], centerTime, -plotTimeRange/2, 
				 plotTimeRange/2, channelNameString);

	// print raw time series to file
	wlog(debugLevel, 2, "    printing raw time series\n");
	ostringstream figName;
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_timeseries_raw";
	string figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                 plot high pass filtered time series               //
	///////////////////////////////////////////////////////////////////////

	// plot high pass filtered time series
	wlog(debugLevel, 2, "    plotting high pass filtered time series\n");
	figureHandle.clear();
	figureHandle.wtimeseries(conditioner.highPassedData(), centerTime, 
				 -plotTimeRange/2, plotTimeRange/2, 
				 channelNameString);

	// print high pass filtered time series to file
	wlog(debugLevel, 2, "    printing high passed time series\n");
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< setprecision(2) << plotTimeRange << "_timeseries_highpassed";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                      plot whitened time series                    //
	///////////////////////////////////////////////////////////////////////

	// plot whitened time series
	wlog(debugLevel, 2, "    plotting whitened time series\n");
	figureHandle.clear();
	figureHandle.wtimeseries(conditioner.whitenedData(), centerTime, 
				 -plotTimeRange/2, plotTimeRange/2, 
				 channelNameString);

	// print whitened time series to file
	wlog(debugLevel, 2, "    printing whitened time series\n");
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_timeseries_whitened";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                         plot raw spectrogram                      //
	///////////////////////////////////////////////////////////////////////
	dble_vect tRange(2);
	tRange[0] = -plotTimeRange/2;
	tRange[1] =  plotTimeRange/2;

	dble_vect qRange(1, mostSignificantQ);

	// plot raw spectrogram
	wlog(debugLevel, 2, "    plotting raw spectrogram\n");
	figureHandle.clear();
	figureHandle.wspectrogram(rawTransform[0], tiling, centerTime, 
		     tRange, plotFrequencyRange, 
		     qRange, plotNormalizedEnergyRange, 
		     plotHorizontalResolution);

	// print autoscaled spectrogram to file
	wlog(debugLevel, 2, "    printing raw spectrogram\n");
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_spectrogram_raw";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");
    
	///////////////////////////////////////////////////////////////////////
	//                      plot whitened spectrogram                    //
	///////////////////////////////////////////////////////////////////////

	// plot whitened spectrogram
	wlog(debugLevel, 2, "    plotting whitened spectrogram\n");
	figureHandle.clear();
	figureHandle.wspectrogram(whitenedTransform[0], tiling,
				  centerTime, tRange, 
				  plotFrequencyRange, 
				  qRange, plotNormalizedEnergyRange, 
				  plotHorizontalResolution);

	// print whitened spectrogram to file
	wlog(debugLevel, 2, "    printing whitened spectrogram\n");
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_spectrogram_whitened";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                     plot autoscaled spectrogram                   //
	///////////////////////////////////////////////////////////////////////

	// plot autoscaled spectrogram
	wlog(debugLevel, 2, "    plotting autoscaled spectrogram\n");
	figureHandle.clear();
	figureHandle.wspectrogram(whitenedTransform[0], tiling, 
				  centerTime, tRange,
				  plotFrequencyRange, qRange, 
				  autoNormalizedEnergyRange, 
				  plotHorizontalResolution);

	// print autoscaled spectrogram to file
	wlog(debugLevel, 2, "    printing autoscaled spectrogram\n");
	figName << eventTimeString << "_" << channelNameString << "_" 
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_spectrogram_autoscaled";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                          plot raw eventgram                       //
	///////////////////////////////////////////////////////////////////////

	// plot raw eventgram
	wlog(debugLevel, 2, "    plotting raw eventgram\n");
	figureHandle.clear();
	if (rawEvents.numberOfChannels()) {
	  figureHandle.weventgram(rawEvents[0], tiling, centerTime, tRange, 
				  plotFrequencyRange, plotDurationInflation, 
				  plotBandwidthInflation, 
				  plotNormalizedEnergyRange);
	}
	// print autoscaled eventgram to file
	wlog(debugLevel, 2, "    printing raw eventgram\n");
	figName << eventTimeString << "_" << channelNameString << "_"
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_eventgram_raw";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                       plot whitened eventgram                     //
	///////////////////////////////////////////////////////////////////////

	// plot whitened eventgram
	wlog(debugLevel, 2, "    plotting whitened eventgram\n");
	figureHandle.clear();
	if (whitenedEvents.numberOfChannels()) {
	  figureHandle.weventgram(whitenedEvents[0], tiling, centerTime, 
				  tRange, plotFrequencyRange, 
				  plotDurationInflation, 
				  plotBandwidthInflation, 
				  plotNormalizedEnergyRange);
	}

	// print whitened eventgram to file
	wlog(debugLevel, 2, "    printing whitened eventgram\n");
	figName << eventTimeString << "_" << channelNameString << "_"
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_eventgram_whitened";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                      plot autoscaled eventgram                    //
	///////////////////////////////////////////////////////////////////////

	// plot autoscaled eventgram
	wlog(debugLevel, 2, "    plotting autoscaled eventgram\n");
	figureHandle.clear();
	if (whitenedEvents.numberOfChannels()) {
	  figureHandle.weventgram(whitenedEvents[0], tiling, centerTime, 
				  tRange, plotFrequencyRange, 
				  plotDurationInflation, 
				  plotBandwidthInflation, 
				  autoNormalizedEnergyRange);
	}

	// print autoscaled eventgram to file
	wlog(debugLevel, 2, "    printing autoscaled eventgram\n");
	figName << eventTimeString << "_" << channelNameString << "_"
		<< fixed << setprecision(2) << plotTimeRange 
		<< "_eventgram_autoscaled";
	figBasePath = outputDirectory + "/" + figName.str();
	figureHandle.wprintfig(figBasePath);
	figName.str("");

	///////////////////////////////////////////////////////////////////////
	//                      add images to html report                    //
	///////////////////////////////////////////////////////////////////////

	// create html link
	wlog(debugLevel, 2, "    linking images\n");
	fprintf(htmlFID, "<a id=\"a_%s_%s_%.2f\"\n", 
		eventTimeString.c_str(), channelNameString.c_str(), 
		plotTimeRange);
	fprintf(htmlFID, "   href=\"%s_%s_%.2f_spectrogram_whitened.png\">\n", 
		eventTimeString.c_str(), channelNameString.c_str(), 
		plotTimeRange);
	fprintf(htmlFID, "<img id=\"img_%s_%s_%.2f\"\n", 
		eventTimeString.c_str(), channelNameString.c_str(), 
		plotTimeRange);
	fprintf(htmlFID, " src=\"%s_%s_%.2f_spectrogram_whitened.thumb.png\"\n", 
		eventTimeString.c_str(), channelNameString.c_str(), 
		plotTimeRange);
	fprintf(htmlFID, "     alt=\"%s_%s_%.2f\" /></a>\n", 
		eventTimeString.c_str(), channelNameString.c_str(), 
		plotTimeRange);

	///////////////////////////////////////////////////////////////////////
	//                        end loop over time ranges                  //
	///////////////////////////////////////////////////////////////////////

	// end loop time ranges
      }

      // end html table row
      fprintf(htmlFID, "<br />\n");
      fprintf(htmlFID, "</div>\n");
      fprintf(htmlFID, "\n");

    } // generateReport

    ///////////////////////////////////////////////////////////////////////////
    //                     end loop over configuration channels              //
    ///////////////////////////////////////////////////////////////////////////

    // end loop over configuration channels
  }

  /////////////////////////////////////////////////////////////////////////////
  //                              close html report                          //
  /////////////////////////////////////////////////////////////////////////////

  if (generateReport) {

    // report status
    wlog(debugLevel, 1, "closing html report\n");

    // end any previously existing section
    if (inSectionFlag == 1) {
      fprintf(htmlFID, "</div>\n");
      fprintf(htmlFID, "\n");
    }

    // end html
    fprintf(htmlFID, "</div>\n");
    fprintf(htmlFID, "\n");
    fprintf(htmlFID, "<div class=\"footer\">\n");
    fprintf(htmlFID, "Created by user %s on %s at %s<br />\n", 
	    getenv("USER"), datestr(0, 29).c_str(), 
	    datestr(0, 13).c_str());
    fprintf(htmlFID, "</div>\n");
    fprintf(htmlFID, "\n");
    fprintf(htmlFID, "</body>\n");
    fprintf(htmlFID, "</html>\n");

    // close html file
    fclose(htmlFID);

  } // generateReport

  /////////////////////////////////////////////////////////////////////////////
  //                          close text summary report                      //
  /////////////////////////////////////////////////////////////////////////////

  // report status
  wlog(debugLevel, 1, "closing text summary");

  // close text summary file
  fclose(textSummaryFID);

  /////////////////////////////////////////////////////////////////////////////
  //                           close xml summary report                      //
  /////////////////////////////////////////////////////////////////////////////
	
  // report status
  wlog(debugLevel, 1, "closing xml summary");

  // end summary table
  fprintf(xmlSummaryFID, "</Stream>\n");
  fprintf(xmlSummaryFID, "</Table>\n");
  fprintf(xmlSummaryFID, "</LIGO_LW>\n");
	
  // close xml summary file
  fclose(xmlSummaryFID);

  /////////////////////////////////////////////////////////////////////////////
  //                                     exit                                //
  /////////////////////////////////////////////////////////////////////////////

  // report completion
  if (debugLevel >= 1) {
    cout << "finished on " << datestr(0, 29) 
	 << " at " << datestr(0, 13) << endl;
  }

  // close all figures
  //figureHandle.close();
  
  // return to calling function
}

//======================================  Search channel constructor.
search_chan::search_chan(void) {
  frameType = "";
  sampleFrequency = 0;
  searchTimeRange = 0;
  searchMaximumEnergyLoss = 0;
  whiteNoiseFalseRate = 0;
  searchWindowDuration = 0;
  alwaysPlotFlag = false;
}
